@@ -31,13 +31,27 @@ FALLBACK_COMMAND_LIST = [
     'load-script', 'dump-cache', 'ab-loop-dump-cache', 'ab-loop-align-cache']
 
 class MPVError(Exception):
+    """An error originating from MPV or due to a problem with MPV."""
     def __init__(self, *args, **kwargs):
         super(MPVError, self).__init__(*args, **kwargs)
 
 class WindowsSocket(threading.Thread):
-    def __init__(self, ipc_socket, callback=None):
+    """
+    Wraps a Windows named pipe in a high-level interface. (Internal)
+    
+    Data is automatically encoded and decoded as JSON. The callback
+    function will be called for each inbound message.
+    """
+    def __init__(self, ipc_socket, callback=None, quit_callback=None):
+        """Create the wrapper.
+
+        *ipc_socket* is the pipe name. (Not including \\\\.\\pipe\\)
+        *callback(json_data)* is the function for recieving events.
+        *quit_callback* is called when the socket connection dies.
+        """
         ipc_socket = "\\\\.\\pipe\\" + ipc_socket
         self.callback = callback
+        self.quit_callback = quit_callback
         
         access = _winapi.GENERIC_READ | _winapi.GENERIC_WRITE
         limit = 5 # Connection may fail at first. Try 5 times.
@@ -59,15 +73,24 @@ class WindowsSocket(threading.Thread):
 
         threading.Thread.__init__(self)
 
-    def stop(self):
+    def stop(self, join=True):
+        """Terminate the thread."""
         if self.socket is not None:
             self.socket.close()
-        self.join()
+        if join:
+            self.join()
 
     def send(self, data):
-        self.socket.send_bytes(json.dumps(data).encode('utf-8') + b'\n')
+        """Send *data* to the pipe, encoded as JSON."""
+        try:
+            self.socket.send_bytes(json.dumps(data).encode('utf-8') + b'\n')
+        except OSError as ex:
+            if len(ex.args) == 1 and ex.args[0] == "handle is closed":
+                raise BrokenPipeError("handle is closed")
+            raise ex
 
     def run(self):
+        """Process pipe events. Do not run this directly. Use *start*."""
         data = b''
         try:
             while True:
@@ -87,12 +110,26 @@ class WindowsSocket(threading.Thread):
                     self.callback(json_data)
                 data = b''
         except EOFError:
-            pass
+            if self.quit_callback:
+                self.quit_callback()
 
 class UnixSocket(threading.Thread):
-    def __init__(self, ipc_socket, callback=None):
+    """
+    Wraps a Unix/Linux socket in a high-level interface. (Internal)
+    
+    Data is automatically encoded and decoded as JSON. The callback
+    function will be called for each inbound message.
+    """
+    def __init__(self, ipc_socket, callback=None, quit_callback=None):
+        """Create the wrapper.
+
+        *ipc_socket* is the path to the socket.
+        *callback(json_data)* is the function for recieving events.
+        *quit_callback* is called when the socket connection dies.
+        """
         self.ipc_socket = ipc_socket
         self.callback = callback
+        self.quit_callback = quit_callback
         self.socket = socket.socket(socket.AF_UNIX)
         self.socket.connect(self.ipc_socket)
 
@@ -101,16 +138,26 @@ class UnixSocket(threading.Thread):
 
         threading.Thread.__init__(self)
 
-    def stop(self):
+    def stop(self, join=True):
+        """Terminate the thread."""
         if self.socket is not None:
-            self.socket.shutdown(socket.SHUT_WR)
-            self.socket.close()
-        self.join()
+            try:
+                self.socket.shutdown(socket.SHUT_WR)
+                self.socket.close()
+                self.socket = None
+            except OSError:
+                pass # Ignore socket close failure.
+        if join:
+            self.join()
 
     def send(self, data):
+        """Send *data* to the socket, encoded as JSON."""
+        if self.socket is None:
+            raise BrokenPipeError("socket is closed")
         self.socket.send(json.dumps(data).encode('utf-8') + b'\n')
 
     def run(self):
+        """Process socket events. Do not run this directly. Use *start*."""
         data = b''
         while True:
             current_data = self.socket.recv(1024)
@@ -128,9 +175,22 @@ class UnixSocket(threading.Thread):
                 json_data = json.loads(item)
                 self.callback(json_data)
             data = b''
+        if self.quit_callback:
+            self.quit_callback()
 
 class MPVProcess:
+    """
+    Manages an MPV process, ensuring the socket or pipe is available. (Internal)
+    """
     def __init__(self, ipc_socket, mpv_location=None, **kwargs):
+        """
+        Create and start the MPV process. Will block until socket/pipe is available.
+
+        *ipc_socket* is the path to the Unix/Linux socket or name of the Windows pipe.
+        *mpv_location* is the path to mpv. If left unset it tries the one in the PATH.
+
+        All other arguments are forwarded to MPV as command-line arguments.
+        """
         if mpv_location is None:
             if os.name == 'nt':
                 mpv_location = "mpv.exe"
@@ -155,13 +215,22 @@ class MPVProcess:
         args.extend("--{0}={1}".format(v[0].replace("_", "-"), self._mpv_fmt(v[1]))
                     for v in kwargs.items())
         self.process = subprocess.Popen(args)
+        ipc_exists = False
         for _ in range(100): # Give MPV 10 seconds to start.
             time.sleep(0.1)
             self.process.poll()
-            if os.path.exists(ipc_socket) or self.process.returncode is not None:
+            if os.path.exists(ipc_socket):
+                ipc_exists = True
+                log.debug("Found MPV socket.")
                 break
+            if self.process.returncode is not None:
+                log.error("MPV failed with returncode {0}.".format(self.process.returncode))
+                break
+        else:
+            self.process.terminate()
+            raise MPVError("MPV start timed out.")
         
-        if not os.path.exists(ipc_socket) or self.process.returncode is not None:
+        if not ipc_exists or self.process.returncode is not None:
             self.process.terminate()
             raise MPVError("MPV not started.")
 
@@ -178,21 +247,32 @@ class MPVProcess:
             return data
 
     def stop(self):
+        """Terminate the process."""
         self.process.terminate()
         if os.name != 'nt' and os.path.exists(self.ipc_socket):
             os.remove(self.ipc_socket)
 
 class MPVInter:
-    def __init__(self, ipc_socket, callback=None):
+    """
+    Low-level interface to MPV. Does NOT manage an mpv process. (Internal)
+    """
+    def __init__(self, ipc_socket, callback=None, quit_callback=None):
+        """Create the wrapper.
+
+        *ipc_socket* is the path to the Unix/Linux socket or name of the Windows pipe.
+        *callback(event_name, data)* is the function for recieving events.
+        *quit_callback* is called when the socket connection to MPV dies.
+        """
         Socket = UnixSocket
         if os.name == 'nt':
             Socket = WindowsSocket
 
         self.callback = callback
+        self.quit_callback = quit_callback
         if self.callback is None:
             self.callback = lambda event, data: None
         
-        self.socket = Socket(ipc_socket, self.event_callback)
+        self.socket = Socket(ipc_socket, self.event_callback, self.quit_callback)
         self.socket.start()
         self.command_id = 1
         self.rid_lock = threading.Lock()
@@ -200,10 +280,12 @@ class MPVInter:
         self.cid_result = {}
         self.cid_wait = {}
     
-    def stop(self):
-        self.socket.stop()
+    def stop(self, join=True):
+        """Terminate the underlying connection."""
+        self.socket.stop(join)
 
     def event_callback(self, data):
+        """Internal callback for recieving events from MPV."""
         if "request_id" in data:
             self.cid_result[data["request_id"]] = data
             self.cid_wait[data["request_id"]].set()
@@ -211,6 +293,14 @@ class MPVInter:
             self.callback(data["event"], data)
     
     def command(self, command, *args):
+        """
+        Issue a command to MPV. Will block until completed or timeout is reached.
+        
+        *command* is the name of the MPV command
+
+        All further arguments are forwarded to the MPV command.
+        Throws TimeoutError if timeout of 120 seconds is reached.
+        """
         self.rid_lock.acquire()
         command_id = self.command_id
         self.command_id += 1
@@ -242,18 +332,29 @@ class MPVInter:
             raise TimeoutError("No response from MPV.")
 
 class EventHandler(threading.Thread):
+    """Event handling thread. (Internal)"""
     def __init__(self):
+        """Create an instance of the thread."""
         self.queue = queue.Queue()
         threading.Thread.__init__(self)
     
     def put_task(self, func, *args):
+        """
+        Put a new task to the thread.
+        
+        *func* is the function to call
+        
+        All further arguments are forwarded to *func*.
+        """
         self.queue.put((func, args))
 
-    def stop(self):
+    def stop(self, join=True):
+        """Terminate the thread."""
         self.queue.put("quit")
-        self.join()
+        self.join(join)
 
     def run(self):
+        """Process socket events. Do not run this directly. Use *start*."""
         while True:
             event = self.queue.get()
             if event == "quit":
@@ -264,12 +365,36 @@ class EventHandler(threading.Thread):
                 log.error("EventHandler caught exception from {0}.".format(event), exc_info=1)
 
 class MPV:
-    def __init__(self, start_mpv=True, ipc_socket=None, mpv_location=None, log_handler=None, loglevel=None, **kwargs):
+    """
+    The main MPV interface class. Use this to control MPV.
+    
+    This will expose all mpv commands as callable methods and all properties.
+    You can set properties and call the commands directly.
+    
+    Please note that if you are using a really old MPV version, a fallback command
+    list is used. Not all commands may actually work when this fallback is used.
+    """
+    def __init__(self, start_mpv=True, ipc_socket=None, mpv_location=None,
+                 log_handler=None, loglevel=None, quit_callback=None, **kwargs):
+        """
+        Create the interface to MPV and process instance.
+
+        *start_mpv* will start an MPV process if true. (Default: True)
+        *ipc_socket* is the path to the Unix/Linux socket or name of Windows pipe. (Default: Random Temp File)
+        *mpv_location* is the location of MPV for *start_mpv*. (Default: Use MPV in PATH)
+        *log_handler(level, prefix, text)* is an optional handler for log events. (Default: Disabled)
+        *loglevel* is the level for log messages. Levels are fatal, error, warn, info, v, debug, trace. (Default: Disabled)
+        *quit_callback* is called when the socket connection to MPV dies.
+
+        All other arguments are forwarded to MPV as command-line arguments if *start_mpv* is used.
+        """
         self.properties = {}
         self.event_bindings = {}
         self.key_bindings = {}
         self.property_bindings = {}
         self.mpv_process = None
+        self.mpv_inter = None
+        self.quit_callback = quit_callback
         self.event_handler = EventHandler()
         self.event_handler.start()
         if ipc_socket is None:
@@ -291,7 +416,7 @@ class MPV:
             else:
                 raise MPVError("MPV process retry limit reached.")
 
-        self.mpv_inter = MPVInter(ipc_socket, self._callback)
+        self.mpv_inter = MPVInter(ipc_socket, self._callback, self._quit_callback)
         self.properties = set(x.replace("-", "_") for x in self.command("get_property", "property-list"))
         try:
             command_list = [x["name"] for x in self.command("get_property", "command-list")]
@@ -318,7 +443,7 @@ class MPV:
         @self.on_event("property-change")
         def event_handler(data):
             if data.get("id") in self.property_bindings:
-                self.event_handler.put_task(self.property_bindings[data["id"]], data["name"], data["data"])
+                self.event_handler.put_task(self.property_bindings[data["id"]], data["name"], data.get("data"))
 
         @self.on_event("client-message")
         def client_message_handler(data):
@@ -326,12 +451,33 @@ class MPV:
             if len(args) == 2 and args[0] == "custom-bind":
                 self.event_handler.put_task(self.key_bindings[args[1]])
 
+    def _quit_callback(self):
+        """
+        Internal handler for quit events.
+        """
+        if self.quit_callback:
+            self.quit_callback()
+        self.terminate(join=False)
+
     def bind_event(self, name, callback):
+        """
+        Bind a callback to an MPV event.
+
+        *name* is the MPV event name.
+        *callback(event_data)* is the function to call.
+        """
         if name not in self.event_bindings:
             self.event_bindings[name] = set()
         self.event_bindings[name].add(callback)
 
     def on_event(self, name):
+        """
+        Decorator to bind a callback to an MPV event.
+
+        @on_event(name)
+        def my_callback(event_data):
+            pass
+        """
         def wrapper(func):
             self.bind_event(name, func)
             return func
@@ -339,15 +485,29 @@ class MPV:
 
     # Added for compatibility.
     def event_callback(self, name):
+        """An alias for on_event to maintain compatibility with python-mpv."""
         return self.on_event(name)
 
     def on_key_press(self, name):
+        """
+        Decorator to bind a callback to an MPV keypress event.
+
+        @on_key_press(key_name)
+        def my_callback():
+            pass
+        """
         def wrapper(func):
             self.bind_key_press(name, func)
             return func
         return wrapper
 
     def bind_key_press(self, name, callback):
+        """
+        Bind a callback to an MPV keypress event.
+
+        *name* is the key symbol.
+        *callback()* is the function to call.
+        """
         self.keybind_lock.acquire()
         keybind_id = self.keybind_id
         self.keybind_id += 1
@@ -362,6 +522,14 @@ class MPV:
             self.enable_section(bind_name)
 
     def bind_property_observer(self, name, callback):
+        """
+        Bind a callback to an MPV property change.
+
+        *name* is the property name.
+        *callback(name, data)* is the function to call.
+
+        Returns a unique observer ID needed to destroy the observer.
+        """
         self.observer_lock.acquire()
         observer_id = self.observer_id
         self.observer_id += 1
@@ -372,19 +540,41 @@ class MPV:
         return observer_id
 
     def unbind_property_observer(self, observer_id):
+        """
+        Remove callback to an MPV property change.
+
+        *observer_id* is the id returned by bind_property_observer.
+        """
         self.command("unobserve_property", observer_id)
         del self.property_bindings[observer_id]
 
     def property_observer(self, name):
+        """
+        Decorator to bind a callback to an MPV property change.
+
+        @property_observer(property_name)
+        def my_callback(name, data):
+            pass
+        """
         def wrapper(func):
             self.bind_property_observer(name, func)
             return func
         return wrapper
     
     def wait_for_property(self, name):
+        """
+        Waits for the value of a property to change.
+
+        *name* is the name of the property.
+        """
         event = threading.Event()
+        first_event = True
         def handler(*_):
-            event.set()
+            nonlocal first_event
+            if first_event == True:
+                first_event = False
+            else:
+                event.set()
         observer_id = self.bind_property_observer(name, handler)
         event.wait()
         self.unbind_property_observer(observer_id)
@@ -400,18 +590,30 @@ class MPV:
                 self.event_handler.put_task(callback, data)
 
     def play(self, url):
+        """Play the specified URL. An alias to loadfile()."""
         self.loadfile(url)
 
     def __del__(self):
         self.terminate()
 
-    def terminate(self):
+    def terminate(self, join=True):
+        """Terminate the connection to MPV and process (if *start_mpv* is used)."""
         if self.mpv_process:
             self.mpv_process.stop()
-        self.mpv_inter.stop()
-        self.event_handler.stop()
+        if self.mpv_inter:
+            self.mpv_inter.stop(join)
+        self.event_handler.stop(join)
 
     def command(self, command, *args):
+        """
+        Send a command to MPV. All commands are bound to the class by default,
+        except JSON IPC specific commands. This may also be useful to retain
+        compatibility with python-mpv, as it does not bind all of the commands. 
+
+        *command* is the command name.
+
+        All further arguments are forwarded to the MPV command. 
+        """
         return self.mpv_inter.command(command, *args)
 
     def __getattr__(self, name):
